/*
	I was helped by looking at the source code of Trevor Blackwell's
	AVR-based unicycle at http://www.tlb.org/eunicycle.html.
*/

#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/sleep.h>

#include <string.h>

#include "adc.h"
#include "defs.h"
#include "fieldsat.h"
#include "mmc.h"
#include "serial.h"
#include "servo.h"

static void init(void);

static void notify_startup(void);

static void notify_happy(void);

static void notify_mmc_failed(void);

static void msleep(unsigned short ms);

static void buzz(unsigned short ms);

static void get_sample_s1(struct fieldsat_sample_s1 *sample);

static void get_sample_s2(struct fieldsat_sample_s2 *sample);

static unsigned short get_timer(void);

static unsigned long get_timestamp(void);

static unsigned char seek_empty(unsigned char mmc);

static void write(const unsigned char *buffer, unsigned int size);

static void write_run_start(void);

static void write_sample_s1(const struct fieldsat_sample_s1 *sample);

static void write_sample_s2(const struct fieldsat_sample_s2 *sample);

static void solar_deploy(unsigned char state);

static volatile unsigned short msleep_compare_count;
static unsigned long timestamp_high = 0;

static unsigned char mmc_is_initialized[2] = { 0, 0 };
static unsigned char mmc_write_has_failed[2] = { 0, 0 };

static unsigned char write_buffer[MMC_SECTOR_SIZE];
static unsigned long write_pos = 0;
static unsigned long sector_num = 0;

static unsigned char solar_has_deployed = 0;
static unsigned char solar_has_locked = 0;

ISR(__vector_default)
{
	#ifdef USE_SERIAL
	serial_puts("Bad interrupt.\n");
	#endif
	asm volatile ("break");
}

/* This interrupt handler is registered and called during msleep. It decrements
   msleep_compare_count, which is monitored by msleep. */
ISR(TIMER1_COMPA_vect)
{
	if (msleep_compare_count > 0)
		msleep_compare_count--;
}

/* This interrupt handler is called when the 16-bit timer overflows. The 16-bit
   variable timestamp_high is used to increase the range of the timer to 32
   bits. */
ISR(TIMER1_OVF_vect)
{
	timestamp_high++;
}

int main(void)
{
	struct fieldsat_sample_s1 s1;
	struct fieldsat_sample_s2 s2;
	unsigned long timestamp, next_timestamp, diff;
	unsigned char s2_count;

	init();

	notify_startup();

#ifdef USE_PARANOID_WRITE
	if (mmc_init(MMC0_CS) && seek_empty(MMC0_CS)) {
#else
	if (mmc_init(MMC0_CS)) {
#endif
		mmc_is_initialized[0] = 1;
		notify_happy();
	} else {
		notify_mmc_failed();
	}

#ifdef USE_PARANOID_WRITE
	if (mmc_init(MMC1_CS) && seek_empty(MMC1_CS)) {
#else
	if (mmc_init(MMC1_CS)) {
#endif
		mmc_is_initialized[1] = 1;
		notify_happy();
	} else {
		notify_mmc_failed();
	}

	write_run_start();

	s2_count = 0;
	next_timestamp = get_timestamp();
	for (;;) {
		get_sample_s1(&s1);
		write_sample_s1(&s1);

		s2_count++;
		if (s2_count >= S2_COUNT) {
			get_sample_s2(&s2);
			write_sample_s2(&s2);
			s2_count = 0;
		}

		/* Deploy the solar panel after the proper time. */
		if (!solar_has_deployed
			&& timestamp_high / (F_CPU >> 16) > SOLAR_DEPLOY_TIME) {
			solar_has_deployed = 1;
			solar_deploy(1);
		}
		/* Undeploy the solar panel only if it has deployed. */
		if (solar_has_deployed && !solar_has_locked
			&& timestamp_high / (F_CPU >> 16) > SOLAR_UNDEPLOY_TIME) {
			solar_has_locked = 1;
			solar_deploy(0);
		}

		#ifdef USE_SAMPLE_RATE_THROTTLE
		/* This is the sample rate throttle. next_timestamp advances in
		   units of SAMPLE_PERIOD, converted from milliseconds to clock
		   cycles. We sleep after each sample long enough to keep the
		   sample rate near SAMPLE_FREQUENCY. */
		/* We want this cycle to end whenever the last cycle ended plus
		   the sample period. */
		next_timestamp += F_CPU / 1000 * SAMPLE_PERIOD;
		timestamp = get_timestamp();
		/* Find out how long we have to wait to get there. */
		diff = next_timestamp - timestamp;
		/* If diff is positive, we need to sleep. If not, we're past
		   where we're supposed to be, so catch up the target
		   timestamp. */
		if ((signed long) diff > 0)
			msleep(diff * 1000 / F_CPU);
		else
			/* If we're already past where we're supposed to be,
			   catch up the target timestamp. */
			next_timestamp = timestamp;
		#endif
	}

	/* Halt the CPU so Avrora knows to quit. */
	asm volatile ("break");

	return 0;
}

static void init(void)
{
	/* Enable interrupts. */
	sei();

	/* Turn on the 16-bit clock with no prescaler. See the ATmega32 data
	   sheet, page 110. */
	TCCR1B = _BV(CS10);
	/* Enable an interrupt when the timer overflows. */
	TIMSK |= _BV(TOIE1);

	/* Set the buzzer pin to output. */
	BUZZER_DDR |= _BV(BUZZER);
	/* Turn off the buzzer by setting its bit. */
	BUZZER_PORT |= _BV(BUZZER);

	serial_init();

	adc_init();

	servo_init();
	/* Our servo is set up such that full clockwise rotation is undeployed
	   solar panels. */
	servo_set_position(SERVO_FULL);
}

/* Display some sort of notification message. */
static void notify_startup(void)
{
	notify_happy();

	#ifdef USE_SERIAL
	serial_puts("FieldSat startup\n");
	serial_puts("AVR Libc " __AVR_LIBC_VERSION_STRING__ "\n");
	#endif
}

/* Give a happy notification, the opposite of an error. */
static void notify_happy(void)
{
	unsigned char i;

	for (i = 0; i < 3; i++) {
		buzz(50);
		msleep(50);
	}
}

static void notify_mmc_failed(void)
{
	buzz(300);
	msleep(300);
	buzz(300);
	msleep(300);
}

/* Sleep for ms milliseconds. This function uses the timer 1 output compare
   register and interrupt. That register and interrupt are reserved for this
   function. */
static void msleep(unsigned short ms)
{
	unsigned long num_cycles;
	unsigned short start_time;
	unsigned char sreg;

	start_time = get_timer();

	num_cycles = F_CPU / 1000 * ms;

	/* Set the output compare register 1 A to the low 16 bits of num_cycles
	   after the current time. See the note on accessing 16-bit registers in
	   get_timer. */
	sreg = SREG;
	cli();
	OCR1A = start_time + num_cycles;
	SREG = sreg;

	/* Loop for the high 16 bits of the number of cycles, plus one. */
	msleep_compare_count = (num_cycles >> 16) + 1;

	/* Enable an interrupt on timer 1 output compare match. */
	TIMSK |= _BV(OCIE1A);

	set_sleep_mode(SLEEP_MODE_IDLE);
	/* msleep_compare_count is modified by the interrupt handler. */
	while (msleep_compare_count > 0)
		sleep_mode();

	/* Disable the interrupt. */
	TIMSK &= ~_BV(OCIE1A);
}

/* Buzz the buzzer for ms milliseconds. */
static void buzz(unsigned short ms)
{
	BUZZER_PORT &= ~_BV(BUZZER);
	msleep(ms);
	BUZZER_PORT |= _BV(BUZZER);
}

static void get_sample_s1(struct fieldsat_sample_s1 *sample)
{
	memset(sample, 0, sizeof(*sample));
	sample->timestamp = get_timestamp();
	/* Read the first sample twice. This gets rid of weird glitches that
	   appear otherwise. I don't know what causes this. */
	sample->acc_low[0] = adc_read_channel(CHANNEL_X_ACC_LOW);
	sample->acc_low[0] = adc_read_channel(CHANNEL_X_ACC_LOW);
	sample->acc_low[1] = adc_read_channel(CHANNEL_Y_ACC_LOW);
	sample->acc_low[2] = adc_read_channel(CHANNEL_Z_ACC_LOW);
	sample->ang_vel[0] = adc_read_channel(CHANNEL_X_ANG_VEL);
	sample->ang_vel[1] = adc_read_channel(CHANNEL_Y_ANG_VEL);
	sample->ang_vel[2] = adc_read_channel(CHANNEL_Z_ANG_VEL);
	sample->solar_luminosity = adc_read_channel(CHANNEL_SOLAR_LUMINOSITY);
	sample->solar_voltage = adc_read_channel(CHANNEL_SOLAR_VOLTAGE);
}

static void get_sample_s2(struct fieldsat_sample_s2 *sample)
{
	memset(sample, 0, sizeof(*sample));
	sample->timestamp = get_timestamp();
	sample->pressure = adc_read_channel(CHANNEL_PRESSURE);
	sample->pressure = adc_read_channel(CHANNEL_PRESSURE);
	sample->temp_int = adc_read_channel(CHANNEL_TEMP_INT);
	sample->temp_ext = adc_read_channel(CHANNEL_TEMP_EXT);
}

/* Return the current value of the 16-bit timer. */
static unsigned short get_timer(void)
{
	unsigned char sreg;
	unsigned short timer;

	/* TCNT1 is a 16-bit register. It has to be read in a special way to
	   prevent the read being interrupted by an interrupt. See the ATmega32
	   data sheet, page 89. */
	/* Save the status register (including the interrupt flag). */
	sreg = SREG;
	/* Disable interrupts. */
	cli();
	/* Read the 16-bit value. */
	timer = TCNT1;
	/* Restore the status register. */
	SREG = sreg;

	return timer;
}

/* Return a 32-bit timestamp. The high 16 bits are the value of timestamp_high
   and the low 16 bits are the current value of the timer. */
static unsigned long get_timestamp(void)
{
	unsigned char sreg;
	unsigned long timestamp;

	sreg = SREG;
	/* Disable interrupts to keep timestamp_high from being modified while
	   we're reading it. Otherwise, there is a race condition. */
	cli();
	timestamp = get_timer();
	if (TIFR & _BV(TOV1)) {
		/* If the TOV1 (Timer 1 overflow 1) bit of the TIFR (timer
		   interrupt flag register) is set, we missed an overflow.
		   Pretend we're at the end of this cycle. This guarantees that
		   the timestamp sequence will be at least non-decreasing. The
		   overflow interrupt will run as soon as we re-enable
		   interrupts. */
		timestamp = 0xFFFF;
	}
	timestamp |= timestamp_high << 16;
	SREG = sreg;

	return timestamp;
}

/* Look for the first all-zero sector, starting at sector_num. */
static unsigned char seek_empty(unsigned char mmc)
{
	unsigned int i;

	for (;;) {
		if (!mmc_read_sector(mmc, write_buffer, sector_num))
			return 0;
		for (i = 0; i < MMC_SECTOR_SIZE; i++) {
			if (write_buffer[i] != 0x00)
				break;
		}
		if (i >= MMC_SECTOR_SIZE)
			return 1;
		sector_num++;
	}
}

static void write(const unsigned char *buffer, unsigned int size)
{
	unsigned int offset;

	#ifdef USE_SERIAL
	serial_write(buffer, size);
	#endif

	offset = 0;
	while (size - offset >= sizeof(write_buffer) - write_pos) {
		memmove(write_buffer + write_pos, buffer + offset, sizeof(write_buffer) - write_pos);
		offset += sizeof(write_buffer) - write_pos;
		if (mmc_is_initialized[0]) {
			if (!mmc_write_sector(MMC0_CS, write_buffer, sector_num)
				&& !mmc_write_has_failed[0]) {
				notify_mmc_failed();
				mmc_write_has_failed[0] = 1;
			}
		}
		if (mmc_is_initialized[1]) {
			if (!mmc_write_sector(MMC1_CS, write_buffer, sector_num)
				&& !mmc_write_has_failed[1]) {
				notify_mmc_failed();
				mmc_write_has_failed[1] = 1;
			}
		}
		write_pos = 0;
		sector_num++;
		/* Beep the first NUM_CONFIRMATION_BEEPS times that
		   NUM_CONFIRMATION_SECTORS are written. */
		if ((sector_num / NUM_CONFIRMATION_SECTORS) <= NUM_CONFIRMATION_BEEPS
			&& sector_num % NUM_CONFIRMATION_SECTORS == 0)
			buzz(20);
	}
	memmove(write_buffer + write_pos, buffer + offset, size - offset);
	write_pos += size - offset;
}

static unsigned char sample_buffer[FIELDSAT_SAMPLE_SIZE];

static void write_run_start(void)
{
	unsigned long timestamp;
	unsigned char i;

	timestamp = get_timestamp();
	i = 0;
	memset(sample_buffer, 0x00, FIELDSAT_SAMPLE_SIZE);
	sample_buffer[i++] = *((unsigned char *) &timestamp);
	sample_buffer[i++] = *((unsigned char *) &timestamp + 1);
	sample_buffer[i++] = *((unsigned char *) &timestamp + 2);
	sample_buffer[i++] = *((unsigned char *) &timestamp + 3);
	sample_buffer[i++] = 'G';
	sample_buffer[i++] = 'O';

	write(sample_buffer, FIELDSAT_SAMPLE_SIZE);
}

static void write_sample_s1(const struct fieldsat_sample_s1 *sample)
{
	unsigned char i;

	i = 0;
	memset(sample_buffer, 0x00, FIELDSAT_SAMPLE_SIZE);
	sample_buffer[i++] = *((unsigned char *) &sample->timestamp);
	sample_buffer[i++] = *((unsigned char *) &sample->timestamp + 1);
	sample_buffer[i++] = *((unsigned char *) &sample->timestamp + 2);
	sample_buffer[i++] = *((unsigned char *) &sample->timestamp + 3);
	sample_buffer[i++] = 'S';
	sample_buffer[i++] = '1';
	sample_buffer[i++] = *((unsigned char *) &sample->acc_low[0]);
	sample_buffer[i++] = *((unsigned char *) &sample->acc_low[0] + 1);
	sample_buffer[i++] = *((unsigned char *) &sample->acc_low[1]);
	sample_buffer[i++] = *((unsigned char *) &sample->acc_low[1] + 1);
	sample_buffer[i++] = *((unsigned char *) &sample->acc_low[2]);
	sample_buffer[i++] = *((unsigned char *) &sample->acc_low[2] + 1);
	sample_buffer[i++] = *((unsigned char *) &sample->ang_vel[0]);
	sample_buffer[i++] = *((unsigned char *) &sample->ang_vel[0] + 1);
	sample_buffer[i++] = *((unsigned char *) &sample->ang_vel[1]);
	sample_buffer[i++] = *((unsigned char *) &sample->ang_vel[1] + 1);
	sample_buffer[i++] = *((unsigned char *) &sample->ang_vel[2]);
	sample_buffer[i++] = *((unsigned char *) &sample->ang_vel[2] + 1);
	sample_buffer[i++] = *((unsigned char *) &sample->solar_luminosity);
	sample_buffer[i++] = *((unsigned char *) &sample->solar_luminosity + 1);
	sample_buffer[i++] = *((unsigned char *) &sample->solar_voltage);
	sample_buffer[i++] = *((unsigned char *) &sample->solar_voltage + 1);

	write(sample_buffer, FIELDSAT_SAMPLE_SIZE);
}

static void write_sample_s2(const struct fieldsat_sample_s2 *sample)
{
	unsigned char i;

	i = 0;
	memset(sample_buffer, 0x00, FIELDSAT_SAMPLE_SIZE);
	sample_buffer[i++] = *((unsigned char *) &sample->timestamp);
	sample_buffer[i++] = *((unsigned char *) &sample->timestamp + 1);
	sample_buffer[i++] = *((unsigned char *) &sample->timestamp + 2);
	sample_buffer[i++] = *((unsigned char *) &sample->timestamp + 3);
	sample_buffer[i++] = 'S';
	sample_buffer[i++] = '2';
	sample_buffer[i++] = *((unsigned char *) &sample->pressure);
	sample_buffer[i++] = *((unsigned char *) &sample->pressure + 1);
	sample_buffer[i++] = *((unsigned char *) &sample->temp_int);
	sample_buffer[i++] = *((unsigned char *) &sample->temp_int + 1);
	sample_buffer[i++] = *((unsigned char *) &sample->temp_ext);
	sample_buffer[i++] = *((unsigned char *) &sample->temp_ext + 1);

	write(sample_buffer, FIELDSAT_SAMPLE_SIZE);
}

static void solar_deploy(unsigned char state)
{
	unsigned long timestamp;
	unsigned char i;

	if (state)
		servo_set_position(SERVO_NONE);
	else
		servo_set_position(SERVO_FULL);

	timestamp = get_timestamp();
	i = 0;
	memset(sample_buffer, 0x00, FIELDSAT_SAMPLE_SIZE);
	sample_buffer[i++] = *((unsigned char *) &timestamp);
	sample_buffer[i++] = *((unsigned char *) &timestamp + 1);
	sample_buffer[i++] = *((unsigned char *) &timestamp + 2);
	sample_buffer[i++] = *((unsigned char *) &timestamp + 3);
	sample_buffer[i++] = 'S';
	sample_buffer[i++] = 'D';
	sample_buffer[i++] = !!state;

	write(sample_buffer, FIELDSAT_SAMPLE_SIZE);
}
